## A wrapper around the python low-level socket api, made by Tyler Clarke, April 5, 2020.
## Feel free to use and distribute under the terms of any freeware license you care to mention

## When using, just extend the ServerListenable class and define the handle_post and handle_get functions (each with the arguments self, data, and connection).
## And the handle_aux function, if you want to.

## The webserver can be run by creating an object of your extended class, and calling the run function on it.

## Enjoy!

import socket
import json
import mimetypes
from socketutils import ServerSocket


class TCPServer:
    def __init__(self,host,port,blocking=True):
        self.server=ServerSocket(host,port)
        self.hooks={"handle":[self.hookcontroller,[self.handle]],
                    "init":[self.hookcontroller,[self.listen]],
                    "mainloop":[self.hookcontroller,[self.run]]}
    def initialize(self):
        self.callhook("init")
    def listen(self,lst=5):
        self.server.listen(lst)
    def run(self):
        connection=self.server.wait_until_connection()
        self.callhook("handle",connection,connection.recieveall())
    def hookcontroller(self,hookdict,hook,*args,**kwargs):
        for x in hookdict[hook][1]:
            x(*args,**kwargs)
    def callhook(self,hook,*args,**kwargs):
        self.hooks[hook][0](self.hooks,hook,*args,**kwargs) ## Call the "hook function" with necessary arguments
    def addhookfunct(self,hook,hookfunct):
        self.hooks[hook][1].append(hookfunct)
    def addhook(self,hook,controller=None):
        if controller==None:
            controller=self.hookcontroller
        self.hooks[hook]=[controller,[]]
    def addExtension(self,extension):
        self.addhookfunct("mainloop",extension.run)
    def delhook(self,hook):
        del self.hooks[hook]
    def delhookfunct(self,hook,index):
        del self.hooks[hook][index]
    def _handle(self,connection):
        data=connection.recieveall()
        self.callhook("handle",connection,data)
    def handle(self,connection,data):
        print("HandleHook Called")
    def start(self):
        self.callhook("init")
        while 1:
            self.callhook("mainloop")



## Classes for every supported protocol.
## No versioning, protocols should deal with that themselves.

## Every protocol should accept ONLY a ClientSocket object in its init function
## Also, these should all have a run function.

class HTTP11:
    statuspairs={200:'OK',404:'Not Found'}
    def makeStatusRow(stcode):
        '''Return an http status row bytes object'''
        return ("HTTP/1.1 "+str(stcode)+" "+HTTP11.statuspairs[stcode]+"\r\n").encode()
    def makeHeaders(data):
        toreturn="".encode()
        for x,y in data.items():
            toreturn+=(str(x)+": "+str(y)+"\r\n").encode()
        return toreturn
    def parseRequest(data):
        toreturn={}
        copy=data.decode()
        copy=copy.replace("\r","")
        toreturn["rawrequesttext"]=str(copy)
        copy=copy.split("\n\n")
        toreturn["content"]=str(copy[1])
        toreturn["rawheaders"]=str(copy[0])
        toreturn["headers"]={}
        cop=copy[0].split("\n")
        for x in cop[1:]:
            p=x.split(": ")
            toreturn["headers"][p[0]]=p[1]
        top=cop[0].split(" ")
        toreturn["reqtype"]=top[0]
        toreturn["reqlocation"]=top[1]
        return toreturn
    def addContent(content):
        return "\r\n".encode()+content.encode()

class HTTPSocket:
    '''The HTTP protocol for ClientSocket transmissions.'''
    def __init__(self,socket):
        self.socket=socket
        self.recvdata=socket.recieve()
    def getHTTPVersion(self,data):
        return data.split("\n")[0].split(" ")[2][5:]
    
class StandardSocket:
    '''The standard protocol in use for a socket with no defined protocol.
Essentially mirrors the socket interface.'''
    def __init__(self,socket):
        self.socket=socket
        self.cache="".encode()
    def sendfile(self,filename,cache=False):
        if cache:
            file=open(filename,"rb")
            self.cache+=file.read()
            file.close()
        else:
            self.socket.s_sendfile(filename)
    def sendtext(self,data,cache=False):
        if cache:
            self.cache+=data.encode()
        else:
            self.socket.s_sendtext(data)
    def sendbytes(self,data,cache=False):
        if cache:
            self.cache+=data
        else:
            self.socket.s_sendbytes(data)
    def emptycache(self,close=True):
        self.socket.s_sendbytes(self.cache)
        if close:
            self.socket.s_close()
    def recieve(self,rcv=1024):
        return self.socket.s_recieve(rcv)
        



import socket


## A wrapper around Sockets, for TCP. Any previously defined compatible protocol is accepted.
class ClientSocket:
    '''A wrapper around the underlying socket API, designed for use with a protocol.
All attribute lookups are redirected to the protocol in use, if you really need
to access them just prefix every attribute with s_ (so s_recieve, s_sendtext, s_sendfile, etc)'''
    def __init__(self,sckt,clidata,blocking):  
        self.clidata=clidata
        self.socket=sckt
##        self.protocol=protocol(self)
##    def __getattribute__(self,attr):
##        if attr[0:2]=="s_":
##            return object.__getattribute__(self,attr[2:])
##        else:
##            return self.s_protocol.__getattribute__(attr)
        if not blocking:
            self.socket.setblocking(False)
    def recieve(self,rcv=1024):
        return self.socket.recv(rcv)
    def recievetext(self,rcv=1024):
        return self.recieve(rcv).decode()
    def recieveall(self,chunks=512):
        '''Recieve all available data, and return it.'''
        data="".encode()
        try:
            while 1:
                data+=self.recieve(chunks)
        except: pass
        return data
    def sendfile(self,filename):
        file=open(filename,"rb")
        self.socket.send(file.read())
        file.close()
    def sendtext(self,data):
        self.socket.send(data.encode())
    def sendbytes(self,data):
        self.socket.send(data)
    def close(self):
        self.socket.shutdown(1)
        self.socket.close()

class ServerSocket:
    def __init__(self,host,port,blocking=False):
        self.host=host
        self.port=port
        self.socket=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        self.socket.bind((host,port))
        self.blocking=False
    def listen(self,amnt=5):
        self.socket.listen(amnt)
    def wait_until_connection(self):
        '''Wait until a connection is recieved, and return a client socket object'''
        self.socket.setblocking(True)
        connection,clientdata=self.socket.accept()
        return ClientSocket(connection,clientdata,self.blocking)
    def get_connection_nonblocking(self):
        '''Get a connection without blocking the program.
Returns None if no connection is found, and returns a client socket otherwise.'''
        self.socket.setblocking(False)
        try:
            connection, clientdata=self.socket.accept()
            return ClientSocket(connection,clientdata,self.blocking)
        except:
            return None



from distutils.core import setup
setup(
  name = 'serverutils',
  packages = ['serverutils'],
  version = '2.1',
  description = 'A python library for webserver development with a simple API',
  author = 'Noman',                   # Type in your name
  author_email = 'plupy44@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/LinuxRocks2000/netutils',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/LinuxRocks2000/netutils/archive/v_01.tar.gz',    # I explain this later on
  keywords = ['socket', 'websockets', 'tcp', 'web', 'server'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'websockets',
  ],
)
